from typing import Optional

from PIL.Image import Image
from ocrd import Workspace, Resolver
from ocrd_modelfactory import page_from_file
from ocrd_models import OcrdFile, OcrdExif
from pathlib import Path
import cv2
import struct
import zlib

from ocrd_models.ocrd_page_generateds import PcGtsType, PageType, MetadataType
from ocrd_utils import pushd_popd


class Document:
    def __init__(self, workspace: Workspace, resolver: Resolver = None, file_group='OCR-D-IMG'):
        self.current_file: OcrdFile = None
        self.workspace: Workspace = workspace
        self.resolver: Resolver = resolver if resolver else Resolver()
        self.file_group: str = file_group

    @property
    def page_ids(self) -> list:
        return self.workspace.mets.physical_pages

    @property
    def file_groups(self) -> list:
        return self.workspace.mets.file_groups

    @classmethod
    def create(cls, directory=None, resolver: Resolver = None) -> 'Document':
        resolver = resolver if resolver else Resolver()
        workspace = resolver.workspace_from_nothing(directory=directory, mets_basename='mets.xml')
        return cls(workspace, resolver)

    @classmethod
    def load(cls, mets_url=None, resolver: Resolver = None) -> 'Document':
        if not mets_url:
            return cls.create(None, resolver)
        resolver = resolver if resolver else Resolver()
        workspace = resolver.workspace_from_url(mets_url, download=True)
        doc = cls(workspace, resolver)
        return doc

    def page_for_id(self, page_id: str, file_group: str = None) -> Optional['Page']:
        page_file = self.file_for_page_id(page_id, file_group)
        if not page_file:
            return None
        pcgts = self.page_for_file(page_file)
        page = pcgts.get_Page()
        image, info, exif = self.workspace.image_from_page(page, page_id)
        return Page(page_id, page_file, pcgts, image, exif)

    def file_for_page_id(self, page_id, file_group=None) -> Optional[OcrdFile]:
        with pushd_popd(self.workspace.directory):
            files = self.workspace.mets.find_files(fileGrp=file_group if file_group else self.file_group,
                                                   pageId=page_id)
            if not files:
                return None
            return self.workspace.download_file(files[0])

    @property
    def directory(self) -> Path:
        return Path(self.workspace.directory)

    def path(self, other) -> Path:
        if isinstance(other, OcrdFile):
            return self.directory / other.local_filename
        elif isinstance(other, Path):
            return self.directory / other
        elif isinstance(other, str):
            return self.directory / other

    def page_for_file(self, page_file: OcrdFile) -> PcGtsType:
        with pushd_popd(self.workspace.directory):
            return page_from_file(page_file)

    def save(self):
        self.workspace.save_mets()

    def add_image(self, image, page_nr, dpi: int = 300) -> 'OcrdFile':
        retval, image_bytes = cv2.imencode(".png", image)
        image_bytes = self._add_dpi_to_png_buffer(image_bytes, dpi)
        file_group = 'OCR-D-IMG'
        file_id = '{}_{:04d}'.format(file_group, page_nr)
        page_id = 'PAGE_{:04d}'.format(page_nr)

        extension = '.png'
        mime_type = 'image/png'
        local_filename = Path(file_group, '%s%s' % (file_id, extension))
        url = Path(self.workspace.directory) / local_filename
        self.current_file = self.workspace.add_file(file_group, ID=file_id, mimetype=mime_type, force=True,
                                                    content=image_bytes, url=str(url),
                                                    local_filename=str(local_filename), pageId=page_id)
        return self.current_file

    @staticmethod
    def _add_dpi_to_png_buffer(image_bytes, dpi=300):
        """
        adds dpi information to a png image

        :param image_bytes:
        :param dpi:
        :return:
        """
        if isinstance(dpi, (int, float)):
            dpi = (dpi, dpi)
        s = image_bytes.tostring()

        # Find start of IDAT chunk
        idat_offset = s.find(b'IDAT') - 4

        # Create our lovely new pHYs chunk - https://www.w3.org/TR/2003/REC-PNG-20031110/#11pHYs
        phys_chunk = b'pHYs' + struct.pack('!IIc', int(dpi[0] / 0.0254), int(dpi[1] / 0.0254), b"\x01")
        phys_chunk = struct.pack('!I', 9) + phys_chunk + struct.pack('!I', zlib.crc32(phys_chunk))

        return s[0:idat_offset] + phys_chunk + s[idat_offset:]

    def add_page(self, page_image, page_nr):
        file_id = '{}_{:04d}'.format(self.file_group, page_nr)
        page_id = 'PAGE_{:04d}'.format(page_nr)
        mime_type = 'image/png'
        basename = file_id + '.png'
        local_filename = self.resolver.download_to_directory(self.workspace.directory, str(page_image), basename,
                                                             subdir=self.file_group, if_exists='overwrite')
        self.workspace.add_file(self.file_group, ID=file_id, mimetype=mime_type, force=True, url=local_filename,
                                local_filename=local_filename, pageId=page_id)


class Page:
    def __init__(self, id_: str, file: OcrdFile, pc_gts: PcGtsType, image: Image, exif: OcrdExif):
        self.id: str = id_
        self.file: OcrdFile = file
        self.pc_gts: PcGtsType = pc_gts
        self.image: Image = image
        self.exif: OcrdExif = exif

    @property
    def page(self) -> PageType:
        return self.pc_gts.get_Page()

    @property
    def meta(self) -> MetadataType:
        return self.pc_gts.get_Metadata()



import io

import cv2
import gi
from PIL.Image import Image
from numpy import array as ndarray

gi.require_version('GdkPixbuf', '2.0')
from gi.repository import GdkPixbuf

__all__ = ['cv_scale', 'cv_to_pixbuf', 'pil_to_pixbuf', 'pil_scale']


def cv_to_pixbuf(cv_image: ndarray) -> GdkPixbuf:
    ret, byte_array = cv2.imencode('.jpg', cv_image)
    return bytes_to_pixbuf(byte_array.tobytes())


def pil_to_pixbuf(im: Image) -> GdkPixbuf:
    bytes_io = io.BytesIO()
    im.save(bytes_io, "PNG" if im.mode in ("LA", "RGBA") else "JPEG")
    return bytes_to_pixbuf(bytes_io.getvalue())


def bytes_to_pixbuf(bytes_: bytes) -> GdkPixbuf:
    loader = GdkPixbuf.PixbufLoader()
    loader.write(bytes_)
    loader.close()
    return loader.get_pixbuf()


def cv_scale(orig: ndarray, w=None, h=None) -> ndarray:
    """
    Scale a cv2 image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    height, width, depth = orig.shape
    new_width, new_height = _calculate_scale(width, height, w, h)
    return cv2.resize(orig, (new_width, new_height))


def pil_scale(orig: Image, w=None, h=None) -> ndarray:
    """
    Scale a Pillow image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    new_width, new_height = _calculate_scale(orig.width, orig.height, w, h)
    thumb = orig.copy()
    thumb.thumbnail((new_width, new_height))
    return thumb


def _calculate_scale(old_width, old_height, new_width: None, new_height: None):
    """
    Calculate scaled image size, while keeping the aspect ratio
    :param old_height:int
    :param old_width:int
    :param new_width:int New width
    :param new_height:int New height
    :return: ndarray
    """
    if new_width and new_height:
        raise RuntimeError('Cant scale both width and height')
    if new_height:
        # noinspection PyUnresolvedReferences
        image_scale = new_height / old_height
    elif new_width:
        # noinspection PyUnresolvedReferences
        image_scale = new_width / old_width
    else:
        image_scale = 1

    return int(old_width * image_scale), int(old_height * image_scale)



import gi

gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk, GLib

import concurrent.futures


class LazyLoadingListStore(Gtk.ListStore):
    init_row = GObject.Property()
    load_row = GObject.Property()
    hash_row = GObject.Property()

    def __init__(self, *column_types, init_row, load_row, hash_row):
        column_types = list(column_types)
        column_types.append(str)
        super().__init__(*column_types)
        self.init_row = init_row
        self.load_row = load_row
        self.hash_row = hash_row
        self.futures = None
        self.row_inserted_handler = self.connect('row-inserted', self._on_row_inserted)
        self.row_changed_handler = self.connect('row-changed', self._on_row_changed)

    def start_loading(self):
        self.futures = {}
        self.submit_all()
        GLib.idle_add(self._collect_workers().__next__, priority=GLib.PRIORITY_DEFAULT_IDLE)

    def submit_all(self):
        pool = concurrent.futures.ThreadPoolExecutor()
        for row in self:
            self._submit_future(row, pool)
        pool.shutdown(wait=False)
        return True

    def _submit_future(self, row, pool=None):
        if self.futures is None:
            return
        own_pool = pool is None
        if own_pool:
            pool = concurrent.futures.ThreadPoolExecutor()

        row_hash = self.hash_row(row)

        if row[-1] != row_hash and row not in self.futures.values():
            future = pool.submit(self.load_row, row[:])
            self.futures[future] = row

        if own_pool:
            pool.shutdown(wait=False)

    def _do_insert(self, position, row):
        if row is not None:
            row = tuple(row) + (None,)
        super()._do_insert(position, row)

    def _on_row_inserted(self, list_store: Gtk.ListStore, _path: Gtk.TreePath, it: Gtk.TreeIter):
        row = list_store[it]
        with self.handler_block(self.row_changed_handler):
            self.init_row(row)
        self._submit_future(row)

    def _on_row_changed(self, list_store: Gtk.ListStore, _path: Gtk.TreePath, it: Gtk.TreeIter):
        row = list_store[it]
        self._submit_future(row)

    def _collect_workers(self):
        for future in concurrent.futures.as_completed(self.futures):
            row = self.futures.pop(future)
            try:
                new_row_data = future.result()
            except Exception as exc:
                print('{} generated an exception: {}'.format(row[0], exc))
            else:
                with self.handler_block(self.row_changed_handler):
                    for i, (old, new) in enumerate(zip(row[:], new_row_data)):
                        if old != new:
                            row[i] = new_row_data[i]
                    row[-1] = self.hash_row(row)
                yield True
        # Futures are finished for now, check back every 50ms if there is something new
        GLib.timeout_add(50, self._collect_workers().__next__, priority=GLib.PRIORITY_DEFAULT_IDLE)
        yield False



# import os
# from time import sleep, clock
# from PIL import Image
# from pathlib import Path
# from ppadb.client import Client as AdbClient
# from retrying import retry
#
#
# class AbstractScanDriver:
#
#     def setup(self):
#         raise NotImplementedError('Please override setup() in your driver')
#
#     def scan(self, timeout=1):
#         raise NotImplementedError('Please override scan() in your driver')
#
#     @staticmethod
#     def verify_image_file(image_file):
#         try:
#             if os.path.getsize(image_file) < 10000:
#                 raise RuntimeError('Too small, thumbnail? ({0}B)'.format(os.path.getsize(image_file)))
#             image = Image.open(image_file)
#             if image.width < 1000 or image.height < 1000:
#                 raise RuntimeError('Looks like a thumbnail ({0}x{1})'.format(image.width, image.height))
#             image.verify()
#             image.close()
#             image = Image.open(image_file)
#             image.thumbnail((1, 1))
#             image.close()
#             return True
#         except Exception as e:
#             return False
#
#
# class AndroidADBDriver(AbstractScanDriver):
#
#     def __init__(self, host="127.0.0.1", port=5037):
#         self.host = host
#         self.port = port
#         self.client = None
#         self.device = None
#
#         self.camera_path = None
#         self.newest_photo = None
#
#     def setup(self):
#         self.client = self._connect()
#         self.device = self._get_device()
#         self.device.input_keyevent('KEYCODE_WAKEUP')
#         self.device.input_keyevent('KEYCODE_MENU')
#         self.device.shell('am start -a android.media.action.STILL_IMAGE_CAMERA')
#         self.camera_path = self.device.shell('echo $EXTERNAL_STORAGE/DCIM/Camera').strip()
#         self.newest_photo = self._get_newest_photo()
#
#     def scan(self, timeout=1):
#         previous_photo = self._get_newest_photo()
#         self.device.input_keyevent('KEYCODE_CAMERA')
#
#         self.newest_photo = self._wait_for_image_file(previous_photo, timeout)
#         local_file, remote_file = self._pull_image_file()
#
#         self._delete_remote_file(remote_file)
#
#         return local_file
#
#     def _wait_for_image_file(self, previous_photo, timeout=1):
#         start = clock()
#         newest_photo = previous_photo
#         while previous_photo == self._get_newest_photo():
#             if clock() - start > timeout:
#                 raise TimeoutError('Waited {0} seconds, still no image', str(timeout))
#             sleep(0.1)
#             newest_photo = self._get_newest_photo()
#
#         return self._get_newest_photo()
#
#     @retry(wait_fixed=50, stop_max_delay=2000)
#     def _pull_image_file(self):
#         local_file = '/tmp/{0}'.format(self.newest_photo)
#         remote_file = '{0}/{1}'.format(self.camera_path, self.newest_photo)
#
#         self.device.pull(remote_file, local_file)
#         self._stat(remote_file)
#
#         remote_size = int(self._stat(remote_file))
#         local_size = int(os.path.getsize(local_file))
#         if remote_size != local_size:
#             # display('{0} - {1}'.format(remote_size, local_size))
#             raise IOError('Image not there yet')
#
#         if not self.verify_image_file(local_file):
#             raise IOError('Image is broken')
#
#         return local_file, remote_file
#
#     def _get_newest_photo(self):
#         return self.device.shell("ls -t1b {0} | head -n1".format(self.camera_path)).strip()
#
#     def _stat(self, file, format_string='%s'):
#         return self.device.shell("stat -c{0} {1}".format(format_string, file)).strip()
#
#     def _delete_remote_file(self, remote_file):
#         self.device.shell('rm {0}'.format(remote_file))
#
#     def _get_device(self):
#         for dev in self.client.devices():
#             if dev.get_state() == 'device':
#                 return dev
#         raise RuntimeError('No device found')
#
#     def _connect(self, tries=1):
#         try:
#             client = AdbClient(host=self.host, port=self.port)
#             client.version()
#             return client
#         except RuntimeError as e:
#             if isinstance(e.__context__, ConnectionRefusedError) and tries > 0:
#                 os.system('adb start-server')
#                 return self._connect(tries - 1)
#             else:
#                 raise e
#
#
# class DummyDriver(AbstractScanDriver):
#     def __init__(self, directory='/home/jk/Projekte/archive-tools/projects/sym-mach/orig'):
#         self.directory = directory
#         self.files = None
#
#     def setup(self):
#         self.files = iter(sorted(Path(self.directory).glob('*.jpg')))
#
#     def scan(self, timeout=1):
#         return next(self.files)



from gi.repository import Gio, GLib


class ActionRegistry:
    def __init__(self, for_widget: Gio.ActionMap):
        self.for_widget = for_widget
        self.actions = {}

    def create(self, name, callback=None, param_type: GLib.Variant = None):
        callback = callback if callback else getattr(self.for_widget, 'on_' + name)
        if param_type is not None:
            action: Gio.SimpleAction = Gio.SimpleAction.new(name, param_type.get_type())
        else:
            action: Gio.SimpleAction = Gio.SimpleAction.new(name)
        action.connect("activate", callback)
        self.for_widget.add_action(action)
        self.actions[name] = action
        return action

    def __getitem__(self, item) -> Gio.SimpleAction:
        return self.actions[item]



# -*- coding: utf-8 -*-
import codecs
import subprocess

import setuptools
from setuptools import setup
from ocrd_browser import __version__

install_requires = open('requirements.txt').read().split('\n')

print("Generating gresource bundle")
subprocess.call(
    [
        "glib-compile-resources",
        "--sourcedir=resources",
        "--target=ocrd_browser/ui.gresource",
        "resources/ocrd-browser.gresource.xml",
    ]
)

setup(
    name='browse-ocrd',
    version=__version__,
    author='Johannes Künsebeck',
    author_email='kuensebeck@googlemail.com',
    description='An extensible viewer for OCRD mets.xml files',
    license='MIT License',
    long_description=codecs.open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/hnesk/browse-ocrd",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    setup_requires=['wheel'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Image Recognition'
    ],
    keywords=['OCR', 'OCRD', 'mets', 'PAGE Xml'],
    entry_points={
        'console_scripts': [
            'browse-ocrd = ocrd_browser.main:main',
        ],
        'ocrd_browser_view': [
            'xml = ocrd_browser.extensions.xmlview:ViewXml',
            'images = ocrd_browser.views:ViewImages',
        ],

    },
    package_data={
        '' : ['*.gresource','*.ui', '*.xml']
    },
)
